import 'dart:async';

import 'package:flutter_live_data/life_cycle.dart';

/// Live Data Structure
class LiveData<T> implements LifeCycleObservable {
  final String? name;
  final T initialValue;
  T _currentValue;
  LifeCycleObserver? _lifeCycleObserver;
  StreamController<T>? streamController;

  LiveData(
    T initValue, {
    this.name,
    StreamController<T>? streamController,
    LifeCycleObserver? lifeCycleObserver,
  })  : this.initialValue = initValue,
        this._currentValue = initValue {
    this.streamController = streamController ?? StreamController<T>();
    if (lifeCycleObserver != null) {
      bind(lifeCycleObserver);
    }
  }

  LiveData<T> bind(LifeCycleObserver lifeCycleObserver) {
    this._lifeCycleObserver = lifeCycleObserver;
    this._lifeCycleObserver?.observeLiveData<T>(this);
    return this;
  }

  Stream<T>? get stream => streamController?.stream;

  LifeCycleObserver? get lifeCycleObserver => _lifeCycleObserver;

  set value(T value) {
    this._currentValue = value;
    if (streamController != null) {
      try {
        streamController?.add(value);
      } catch (e) {
        close();
      }
    }
  }

  T get value => this._currentValue;

  void mutate(void Function(T) setter) {
    setter(this._currentValue);
    value = this._currentValue;
  }

  void transform(T Function(T) setter) {
    value = setter(this._currentValue);
  }

  void close() {
    streamController?.close();
    streamController = null;
  }

  StreamSubscription<T>? subscribe(
    void onData(T event), {
    Function? onError,
    void onDone()?,
    bool? cancelOnError,
  }) {
    return stream?.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }
}
